; DtZ's (http://XEPb.ru/dtz) SD card interface Loader  for 
; Searle Grant's (http://searle.hostei.com/grant/cpm/index.html) SBC
;
;

; SD/MMC RELATED DEFINES:

SD_PORT        equ     $8

; NOTE:
; ***WRITE**
; BIT 0 - SCK, normally low
; BIT 1 - DATA, normally data or HIGH while reading
; BIT 2 - CS, chip selecte
; ***READ**
; BIT 1 - INVERTED data. If you use positive data, comment XOR 1 in 
; sd_read_byte_read_bit:


; Definitions for MMC/SDC command 
SD_CMD0		equ	$40 ; GO_IDLE_STATE  Resets the multimedia card
SD_CMD1		equ	$41	; SEND_OP_COND (MMC)  Activates the card's initialization process
;SD_ACMD41	equ	(0xC0+41)	; SEND_OP_COND (SDC)
SD_CMD8		equ	$48	; SEND_IF_COND
SD_CMD9		equ	$49	; SEND_CSD  : Ask the selected card to send its card specific data
SD_CMD10	equ	$4a	; SEND_CID : Ask the selected card to send its card identification
;SD_CMD12	equ	$4c	; STOP_TRANSMISSION
SD_CMD13	equ	$4d	; SD_STATUS (SDC) Ask the selected card to send its status register

SD_CMD16	equ	$50	; SET_BLOCKLEN Select a block length (in bytes) for all following block commands (Read:between 1-512 and Write:only 512)
SD_CMD17	equ	$51	; READ_SINGLE_BLOCK
;SD_ACMD23	(0xC0+23)	; SET_WR_BLK_ERASE_COUNT (SDC) 
SD_CMD24	equ	$58	; WRITE_BLOCK Writes a block of the size selected by CMD16, the start address must be alligned on a sector boundry, the block length is always 512 bytes
SD_CMD27	equ	$5b	; Programming of the programmable bits of the CSD
SD_CMD41	equ	$69
SD_CMD55	equ	$77	;
SD_CMD58	equ	$7a	;



BLOCKSIZE       equ     512     ; SD/MMC block size (bytes)
; I suppose you are in big trouble if it's not 512

cs_high:
	ld	a,6	; CS = 1, DATA = 1, CLK = 0
	out	(SD_PORT),a
	ret

cs_low:
	ld	a,2	; CS = 0, DATA = 1, CLK = 0
	out	(SD_PORT),a
	ret

;********************************************* READING BYTE *****************************************
sd_read_byte:
	push	bc
	ld	b,8	; will read 8 bits
	xor	a	; Initially we have 0
	ld	c,a	; we will keep it to c

sd_read_byte_read_bit:

	ld	a,c	; c = c*2 (or can we rotate C left?)
	add	a,a
	ld	c,a

	ld	a,3	; CS= 0, data=1, clock=1
	out 	(SD_PORT),a ;Should be read at Falling edge, but this is a rising one?
	in	a,(SD_PORT)
;	push	af
;	call	outbytehex
;	pop	af
	and	1	; only	lower bit is significant
	xor	1	; our input is inverted
	add	a,c	
	ld	c,a

	ld	a,2	; CS=0, data=1,clock=0
	out     (SD_PORT),a
	djnz	sd_read_byte_read_bit

	ld	a,c	; helding result in a
	push	af
;	call	outstringinline
;	db	"Read byte: ",0
;	pop	af
;	push	af
;	call	outbytehex

	pop	af
	
	pop	bc
	ret

;*********************** SEND COMMAND, WAIT FOR NON-FF  REPLY **************/
sd_write_cmd_and_get_result:
	call	sd_write_cmd
sd_write_cmd_and_get_result_preloop:
	ld	b,16
sd_write_cmd_and_get_result_loop:	; Normally this loop executes at least twice,
	push	bc			; FF for the first time
	call	sd_read_byte
	inc	a				; was FF, became 0 ; was command, became command +1
	jp	nz, sd_write_cmd_and_get_result_done
	pop	bc
	djnz	sd_write_cmd_and_get_result_loop
	dec	a
	ret		;no way - failed
	
sd_write_cmd_and_get_result_done:
	pop	bc		; remove from stack
	dec	a
	ret
;********************* READ BYTES FOR NON-FF REPLY *********************/
sd_wait_for_non_ff:
	push 	bc
	call	sd_write_cmd_and_get_result_preloop
	pop	bc
	ret


;*********************** SEND COMMAND WITH ARGS, WAIT FOR NON-FF REPLY **************
sd_write_cmd_with_args_and_get_result:
	call	sd_write_cmd_with_args
	jp	sd_write_cmd_and_get_result_preloop


;*************************** WRITING 6-BYTE COMMAND **********************************
;  Sends a command with all the parameters = 00 and checksum = $95. Destroys AF.
; Called from sd_init_:
sd_write_cmd:
	ifdef	DTZDEBUG
	push	af
	call	outstringinline
	db	"Writing Command: ",0
	pop	af
	push	af
	call	outbytehex
	call	outstringinline
	db $0D,$0A,0
	pop	af
	endif

	call	sd_write_byte
	xor	a
	call	sd_write_byte
	xor	a
	call	sd_write_byte
	xor	a
	call	sd_write_byte
	xor	a
	call	sd_write_byte
	ld	a,$95			; $95 is only needed when the CARD INIT is being performed			
	call	sd_write_byte		; then this byte is ignored.

	ifdef	DTZDEBUG
	call	outstringinline
	db $0D,$0A,0
	endif

	ret

;  Sends a command with parameters = 00 and checksum = $95. Destroys AF.
; Called from sd_init_:
sd_write_cmd_with_args:		; args	bc,de
	push	de		; third and forth args Low Looow
	push	bc		; first and second args Hii Hi
;	push	af
;	call	outstringinline
;	db	"Writing Command with args: ",0
;	pop	af
;	push	af
;	call	outbytehex
;	call	outstringinline
;	db $0D,$0A,0
;	pop	af		; command itself
	call	sd_write_byte
	pop	bc
	ld	a,b		; b
	push	bc
	call	sd_write_byte
	pop	bc
	ld	a,c		; c
	call	sd_write_byte
	pop	de
	ld	a,d		; d
	push	de
	call	sd_write_byte
	pop	de
	ld	a,e		; e
	call	sd_write_byte
	ld	a,$95			; $95 is only needed when the CARD INIT is being performed			
	call	sd_write_byte		; then this byte is ignored.
	ret


; Called by sd_write_command
sd_write_byte:
	push	bc
;	push	af
;	call	outstringinline
;	db	"WB: ",0
;	pop	af

	ifdef	DTZDEBUG
	push	af
	call	outbytehex
	pop	af
	endif

	ld	b,8
sd_write_command_bit_loop:
;	push	af
;	ld	a,2
;	out	(SD_PORT),a	; CS 0 Data 1 Clk 0 , Falling edge  /*** THINK ***/
;	pop	af
	add	a,a
	push	af
	ld	a,0		; will XOR clear C flag?
	adc	a,0		; will add C = 1 for bit 1 and 0 for bit 0
	add	a,a		; if was c -> a==2
	out     (SD_PORT),a   ; Data on data line, Clock low
	inc	a		; Clk to high
	out 	(SD_PORT),a	; 
;	push	af
;	call	outbytehex
;	pop	af
        dec	a		; clock to low, data still on data bus, but noone cares! 
	out 	(SD_PORT),a
	pop	af
	djnz	sd_write_command_bit_loop
;	call	outstringinline
;	db	"..Done",$0A,$0D,0
	pop	bc	; pushed after sd_write_byte
	ret

mmc_send_ticks_cs_high:
	push	bc
	ld	a,7
	out 	(SD_PORT),a	; cs-high,data-high,clock-up

	ifdef	DTZDEBUG
	ld	a,'^'
	rst	8
	endif

	ld	a,6
	out     (SD_PORT),a	; cs-high, data-high, clock-down

	ifdef	DTZDEBUG
	ld	a,'v'
	rst	8
	endif

	pop	bc
	djnz	mmc_send_ticks_cs_high
	ret

; Reads 60h 200h-bytes sectors (3000h bytes) into (HL)
; for the 64k, max sector number is 20000h
; for the 128k, max sector is 40000hs
;sd_read_60h_sectors:
;	ld	b,0
;	push	bc
;sd_read_60h_sectors_loop
;	ld	a,SD_CMD17
;	call	sd_write_cmd_and_get_result
;	ld	a,$0D
;	rst	8
;	ld	a,$0A
;	rst	8
;	ld	bc,$210
;readblockloop:
;	push	bc
;	call	sd_read_byte
;	pop	bc
;	dec	bc
;	xor	a
;	or	b
;	or	c	; I suppose a will be 0 only if b and c are 0 
;	jp	nz,readblockloop

; ********************************* External Entry Point ***************************************

sd_init:      
	call	outstringinline
	db	"Initing MMC",$0D,$0A,0
        call   cs_high                 ; set cs high
	ld	b,80
	call	mmc_send_ticks_cs_high
	call	cs_low
;	ld	a,MMC_GO_IDLE_STATE
	ld	a,SD_CMD0

	call	sd_write_cmd

	ifdef	DTZDEBUG
	call	outstringinline
	db	"SD_CMD0 Command written",$0A,$0D,0
	endif

	ld	b,16		; wait for 01 response for 16 byte ticks - should be more then enought
sd_init_wait_init:
	call	sd_read_byte
	dec	a
	jp	z,sd_init_spi_done	; got 01 response, cool! (and 00 result should be OK,too,theoretically. Or not?)
	cp	254			; compare with FF
	jp	nz,sd_init		; got smth - not FF - reiniting again - do not forget to loop this jump
	djnz	sd_init_wait_init	
	call	outstringinline
	db	"Unable to SD_CMD0",$0A,$0D,0
	jp	sd_init_exit
sd_init_spi_done:

	ifdef	DTZDEBUG
	call	outstringinline
	db	"**Entered SPI mode**",$0A,$0D,"Lets reset it",$0A,$0D,0 ; or at least I think we entered
	endif

;http://elm-chan.org/docs/mmc/mmc_e.html

	ld	b,$FF
sd_init_init_loop:
	push	bc
	ld	a,SD_CMD1
	call	sd_write_cmd_and_get_result
	jr	z,sd_init_init_done
	pop	bc
	djnz	sd_init_init_loop
	jp	sd_not_cmd1
sd_init_init_done:
	pop	bc	; get bc out of stacl after sd_init_init_loop:

	ifdef	DTZDEBUG
	call	outstringinline
	db	"CMD1 ok",$0D,$0A,0
	endif

;	Let's read CMD58/GetCCS - some voltage stuff
	ld	a,SD_CMD58		; 7A
	call	sd_write_cmd_and_get_result ; and lets get 4 bytes of data
	call	sd_read_byte
	call	sd_read_byte
	call	sd_read_byte
	call	sd_read_byte

;	Setting sector size
        ld      bc,0            ; high 2 bytes
        ld      de,BLOCKSIZE        ; Means 512, or I bet you are in trouble
        ld      a,SD_CMD16      ; Set sector size to 512 bytes
        call    sd_write_cmd_with_args_and_get_result
        or      a
        jp      nz,sd_not_sector ; This command may fail on some old cards, which default sector size is 512, however
    ; No, that was a speach of SD HC cards. Or not? Do not care, this command works ;-)

; **************** CMD10 *********************
; Prining some card info - we do not care, but it is interesting
;
;

	ld      a,SD_CMD10
	call	sd_write_cmd_with_args_and_get_result
	or	a
	jp	nz,sd_init_cant_get_cmd10_warning

	ld	b,16
sd_init_get_cmd10_wait_FE:		; waiting for data token
	call	sd_read_byte
	cp	$FE
	jp	z,sd_init_get_cmd10_read_data
	djnz	sd_init_get_cmd10_wait_FE
	jp	sd_init_cant_get_cmd10_fe_warning

sd_init_get_cmd10_read_data:

	ifdef	DTZDEBUG
	call	outstringinline
	db	"**Got it*",$0D,$0A,0
	endif

	ld	b,3	;$16 bytes + 2 crc ; first 3 bytes is manufacturer ID + etc ID
sd_init_get_cmd10_loop_0:
	call	sd_read_byte
	djnz	sd_init_get_cmd10_loop_0
	call	outstringinline
	db	"Card ID:",0
	ld	b,5	;next 5 bytes - ascii name of catd
sd_init_get_cmd10_loop_1: ;	printing ASCII card id
	call	sd_read_byte
	rst	8
	djnz	sd_init_get_cmd10_loop_1
	ld	a,$0D
	rst	8
	ld	a,$0A
	rst	8
	ld	b,10	; read to the end of data
sd_init_get_cmd10_loop_2:
	call	sd_read_byte
	djnz	sd_init_get_cmd10_loop_2
	jp	init_mmc_go_cmd10_done

sd_init_cant_get_cmd10_warning: ; just a warning, but very bad
	push	af
	call	outstringinline
	db	"Warning: cannot get info via CMD10: ",0
	pop	af
	call	outbytehex
	call	outstringinline
	db	" instead of 0",$0D,$0A,0
	jp	init_mmc_go_cmd10_done

sd_init_cant_get_cmd10_fe_warning: ; just a warning, but very bad
	call	outstringinline
	db	"Warning: cant get data token for CMD10",$0D,$0A,0
init_mmc_go_cmd10_done:

; Prining some card info - we do not care, but it is interesting
; Card size somehow is fetched by this code
;init_mmc_go_cmd9:
;	ld      a,SD_CMD9
;	call	sd_write_cmd_with_args_and_get_result
;	ld	b,20	; Seems to have 20 significant bytes
;sd_init_get_cmd9_loop_0:
;	call	sd_read_byte
;	djnz	sd_init_get_cmd9_loop_0


; Now we have to read $3000 bytes = $200 bytes sectors.
; That means we have to read 
; b  c  d  e for sd_write_cmd_with_args_and_get_result
;00 00 00 00 for the sector 0
;00 00 02 00 for the sector 1
;00 00 3E 00 for last sector
load_cpm:
	ld	d,$FE	; -2 ; will be 0 for the first iteration
	ld	hl,$C000	; CP/M Load address
load_cpm_loop:	

readblock:
	ld	bc,0
	ld	e,0
	inc	d
	inc	d
;	ld	a,$4		; For now, lets's try to load only two sectors
	ld	a,$40		; loop from 0 to $3e
	cp	d
	jp	z,load_cpm_done

	ld	a,SD_CMD17
	push	de		; Push sector counter
	push	hl		; Push memory pointer

	call	sd_write_cmd_with_args_and_get_result

	ifdef	DTZDEBUG
	push	af
	ld	a,$0D
	rst	8
	ld	a,$0A
	rst	8
	pop	af
	call	outbytehex
	ld	a,$0D
	rst	8
	ld	a,$0A
	rst	8
	endif

		; Should wait for data token started with FE
		; and skip some bytes

	call	sd_wait_for_non_ff
	cp	$FE
	jp	nz, load_cpm_cannot_read_block
;	call	outstringinline
;	db	"Got FE, reading block",$0D,$0A,0
	ld	bc,$200 ; normally $200 , and then read a CRC

;	call	outstringinline
;	db	"HL:",0
	pop	hl		; pop memptr
;	push	hl
;	ld	a,h
;	call	outbytehex
;	ld	a,l
;	call	outbytehex
;	call	outstringinline
;	db	$0D,$0A,0
;	pop	hl		; pop mempt
readblockloop:
	push	bc		; Push byte-inside-sector counter
	push	hl
	call	sd_read_byte
;	push	af
;	call	outbytehex
;	pop	af
	pop	hl
	ld	(hl),a		; Move loaded byte to memory
	inc	hl		; memory counter++
	pop	bc		; Pop byte-inside-sector counter
	dec	bc
	xor	a
	or	b
	or	c	; I suppose A will be 0 only if B and C are 0 
	jp	nz,readblockloop
	push	hl		; push memory pointer to stack	

	ld	a,'.'
	rst	8
;	ld	a,$0D
;	rst	8
;	ld	a,$0A
;	rst	8
	call	sd_read_byte	; Checksum
	call	sd_read_byte	; Checking checksum is generally a good idea

	pop	hl	; pop memory counter
	pop	de	; Pop sector counter
	jp	load_cpm_loop
	
load_cpm_done:
	call	outstringinline
	db	$0D,$0A,"OS Loaded",$0D,$0A,0


;	ld	hl,$D000
;dump:
;	ld	a,h
;	call	outbytehex
;	ld	a,l
;	call	outbytehex
;	ld	a,':'
;	rst	8
;	ld	b,10h
;	push	bc
;dump_loop1:
;	ld	a,(hl)
;	inc	hl
;	call	outbytehex
;	djnz	dump_loop1

;	pop	bc
;        djnz	dump


; Actually run CP/M
	ld	a,(primaryIO)	; PrimaryIO from monitor
	push	af	; should be pushed to stack

	ifdef	DTZDEBUG
	ld	a,($fffe)
	call	outbytehex
	ld	a,($ffff)
	call	outbytehex
	endif
	ld	hl,($fffe)
	jp	(hl)

; Global exit point
sd_init_exit:
	ret

; Show specific error messages

load_cpm_cannot_read_block:
	pop	hl	; remove pointer from stack
	pop	hl	; remove sector counter from stack

	push	af
	call	outstringinline
	db	"Cannot read block: error ",0
	pop	af
	call	outbytehex
	call	outstringinline
	db	$0D,$0A,0
	jp	sd_init_exit

;sd_no_cmd8:
;	call	outstringinline
;	db	"Cannot send CMD8 - voltage problem?",$0D,$0A,0
;	jp	sd_init_exit

sd_not_cmd1:
	call	outstringinline
	db	"Cannot send CMD1",$0D,$0A,0
	jp	sd_init_exit

sd_not_sector:
	call	outstringinline
	db	"Unable to set 512b sector",$0D,$0A,0
	jp	sd_init_exit

;sd_not_cmd41:
;	call	outstringinline
;	db	"CMD41 failed",$0D,$0A,0
;	jp	sd_init_exit

;sd_not_second55:
;	call	outstringinline
;	db	"Second 55 command failed",$0D,$0A,0
;	jp	sd_init_exit

;sd_not_sd_card:
;	call	outstringinline
;	db	"SD card not ready (CMD41)",$0D,$0A,0
;	jp	sd_init_exit



;sd_toggle_cs:
;	call	cs_high
;	ld	b,255
;sd_toggle_cs_loop:
;	djnz	sd_toggle_cs_loop
;	call	cs_low
;	ret
